import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer


class ES(Optimizer):
    """Evolution Strategies (ES).

    This is the **abstract** class for all `ES` classes. Please use any of its instantiated subclasses to
    optimize the black-box problem at hand.

    .. note:: `ES` are a well-established family of randomized **population-based** search algorithms, proposed by two
       German computer scientists Ingo Rechenberg and Hans-Paul Schwefel (two recipients of `IEEE Evolutionary
       Computation Pioneer Award 2002 <https://tinyurl.com/456as566>`_). One key property of `ES` is **adaptability of
       strategy parameters**, which generally can *significantly* accelerate the (local) convergence rate. Recently, the
       **theoretical foundation** of its most representative (modern) version called **CMA-ES** has been well built on
       the `Information-Geometric Optimization (IGO) <https://www.jmlr.org/papers/v18/14-467.html>`_ framework via
       **invariance** principles (inspired by `NES <https://jmlr.org/papers/v15/wierstra14a.html>`_).

       According to the latest `Nature <https://www.nature.com/articles/nature14544.>`_ review, **"the CMA-ES algorithm
       is widely regarded as the state of the art in numerical optimization"**.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'n_individuals' - number of offspring/descendants, aka offspring population size (`int`),
                * 'n_parents'     - number of parents/ancestors, aka parental population size (`int`),
                * 'mean'          - initial (starting) point (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'sigma'         - initial global step-size, aka mutation strength (`float`).

    Attributes
    ----------
    mean          : `array_like`
                    initial (starting) point, aka mean of Gaussian search/sampling/mutation distribution.
    n_individuals : `int`
                    number of offspring/descendants, aka offspring population size.
    n_parents     : `int`
                    number of parents/ancestors, aka parental population size.
    sigma         : `float`
                    global step-size, aka mutation strength (i.e., overall std of Gaussian search distribution).

    Methods
    -------

    References
    ----------
    https://homepages.fhv.at/hgb/downloads/ES-Is-Not-Gradient-Follower.pdf

    Ollivier, Y., Arnold, L., Auger, A. and Hansen, N., 2017.
    Information-geometric optimization algorithms: A unifying picture via invariance principles.
    Journal of Machine Learning Research, 18(18), pp.1-65.
    https://www.jmlr.org/papers/v18/14-467.html

    https://blog.otoro.net/2017/10/29/visual-evolution-strategies/

    Hansen, N., Arnold, D.V. and Auger, A., 2015.
    Evolution strategies.
    In Springer Handbook of Computational Intelligence (pp. 871-898).
    Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007%2F978-3-662-43505-2_44

    Bäck, T., Foussette, C., & Krause, P. (2013).
    Contemporary evolution strategies.
    Berlin: Springer.
    https://link.springer.com/book/10.1007/978-3-642-40137-4

    http://www.scholarpedia.org/article/Evolution_strategies

    Beyer, H.G. and Schwefel, H.P., 2002.
    Evolution strategies–A comprehensive introduction.
    Natural Computing, 1(1), pp.3-52.
    https://link.springer.com/article/10.1023/A:1015059928466

    Rechenberg, I., 2000.
    Case studies in evolutionary experimentation and computation.
    Computer Methods in Applied Mechanics and Engineering, 186(2-4), pp.125-140.
    https://www.sciencedirect.com/science/article/abs/pii/S0045782599003813

    Rechenberg, I., 1989.
    Evolution strategy: Nature’s way of optimization.
    In Optimization: Methods and Applications, Possibilities and Limitations (pp. 106-126).
    Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007/978-3-642-83814-9_6

    Schwefel, H.P., 1988.
    Collective intelligence in evolving systems.
    In Ecodynamics (pp. 95-100). Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007/978-3-642-73953-8_8

    Schwefel, H.P., 1984.
    Evolution strategies: A family of non-linear optimization techniques based on imitating
    some principles of organic evolution.
    Annals of Operations Research, 1(2), pp.165-167.
    https://link.springer.com/article/10.1007/BF01876146

    Rechenberg, I., 1984.
    The evolution strategy. A mathematical model of darwinian evolution.
    In Synergetics—from Microscopic to Macroscopic Order (pp. 122-132). Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007/978-3-642-69540-7_13
    """
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        if self.n_individuals is None:  # number of offspring (λ: lambda), offspring population size
            self.n_individuals = 4 + int(3*np.log(self.ndim_problem))  # only for small populations setting
        assert self.n_individuals > 0, f'`self.n_individuals` = {self.n_individuals}, but should > 0.'
        if self.n_parents is None:  # number of parents (μ: mu), parental population size
            self.n_parents = int(self.n_individuals/2)
            if self.n_parents > 1:
                self._w, self._mu_eff = self._compute_weights()
                self._e_chi = np.sqrt(self.ndim_problem)*(  # E[||N(0,I)||]: expectation of chi distribution
                    1.0 - 1.0/(4.0*self.ndim_problem) + 1.0/(21.0*np.square(self.ndim_problem)))
        assert self.n_parents > 0, f'`self.n_parents` = {self.n_parents}, but should > 0.'
        self.mean = options.get('mean')  # mean of Gaussian search/sampling/mutation distribution
        if self.mean is None:  # `mean` overwrites `x` if both are set
            self.mean = options.get('x')
        self.sigma = options.get('sigma')  # global step-size (σ), mutation strength
        assert self.sigma > 0, f'`self.sigma` = {self.sigma}, but should > 0.'
        self.lr_mean = options.get('lr_mean')  # learning rate of mean update
        assert self.lr_mean is None or self.lr_mean > 0,\
            f'`self.lr_mean` = {self.lr_mean}, but should > 0.'
        self.lr_sigma = options.get('lr_sigma')  # learning rate of sigma update
        assert self.lr_sigma is None or self.lr_sigma > 0,\
            f'`self.lr_sigma` = {self.lr_sigma}, but should > 0.'
        self._printed_evaluations = self.n_function_evaluations
        # set options for *restart*
        self._n_restart = 0  # only for restart
        self._n_generations = 0  # number of generations
        self._list_generations = []  # list of number of generations for all restarts
        self._list_fitness = [self.best_so_far_y]  # only for restart
        self._list_initial_mean = []  # list of mean for each restart
        self.sigma_threshold = options.get('sigma_threshold', 1e-12)  # stopping threshold of sigma
        self.stagnation = options.get('stagnation', int(10 + np.ceil(30*self.ndim_problem/self.n_individuals)))
        self.fitness_diff = options.get('fitness_diff', 1e-12)  # stopping threshold of fitness difference
        self._sigma_bak = np.copy(self.sigma)  # only for restart

    def _compute_weights(self):
        # unify these settings in the base class for consistency and simplicity
        w_base, w = np.log((self.n_individuals + 1.0)/2.0), np.log(np.arange(self.n_parents) + 1.0)
        w = (w_base - w)/(self.n_parents*w_base - np.sum(w))
        mu_eff = 1.0/np.sum(np.square(w))  # μ_eff (μ_w)
        return w, mu_eff

    def initialize(self):
        raise NotImplementedError

    def iterate(self):
        raise NotImplementedError

    def _initialize_mean(self, is_restart=False):
        if is_restart or (self.mean is None):
            mean = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary)
        else:
            mean = np.copy(self.mean)
        self.mean = np.copy(mean)
        return mean

    def _print_verbose_info(self, fitness, y, is_print=False):
        if y is not None:
            if self.saving_fitness:
                if not np.isscalar(y):
                    fitness.extend(y)
                else:
                    fitness.append(y)
        if self.verbose:
            is_verbose = self._printed_evaluations != self.n_function_evaluations  # to avoid repeated printing
            is_verbose_1 = (not self._n_generations % self.verbose) and is_verbose
            is_verbose_2 = self.termination_signal > 0 and is_verbose
            is_verbose_3 = is_print and is_verbose
            if is_verbose_1 or is_verbose_2 or is_verbose_3:
                info = '  * Generation {:d}: best_so_far_y {:7.5e}, min(y) {:7.5e} & Evaluations {:d}'
                print(info.format(self._n_generations, self.best_so_far_y, np.min(y), self.n_function_evaluations))
                self._printed_evaluations = self.n_function_evaluations

    def restart_reinitialize(self, y):
        min_y = np.min(y)
        if min_y < self._list_fitness[-1]:
            self._list_fitness.append(min_y)
        else:
            self._list_fitness.append(self._list_fitness[-1])
        is_restart_1, is_restart_2 = self.sigma < self.sigma_threshold, False
        if len(self._list_fitness) >= self.stagnation:
            is_restart_2 = (self._list_fitness[-self.stagnation] - self._list_fitness[-1]) < self.fitness_diff
        is_restart = bool(is_restart_1) or bool(is_restart_2)
        if is_restart:
            self._print_verbose_info([], y, True)
            if self.verbose:
                print(' ....... *** restart *** .......')
            self._n_restart += 1
            self._list_generations.append(self._n_generations)  # for each restart
            self._n_generations = 0
            self._list_fitness = [np.Inf]
            self.sigma = np.copy(self._sigma_bak)
            self.n_individuals *= 2
            self.n_parents = int(self.n_individuals/2)
            if self.n_parents > 1:
                self._w, self._mu_eff = self._compute_weights()
        return is_restart

    def _collect(self, fitness=None, y=None, mean=None):
        self._print_verbose_info(fitness, y)
        results = Optimizer._collect(self, fitness)
        results['mean'] = mean  # final mean of search distribution
        results['initial_mean'] = self.mean  # initial mean of search distribution
        results['_list_initial_mean'] = self._list_initial_mean  # list of initial mean for each restart
        # by default, do NOT save covariance matrix of search distribution in order to save memory,
        # owing to its *quadratic* space complexity
        results['sigma'] = self.sigma  # only global step-size of search distribution
        results['_n_restart'] = self._n_restart  # number of restart
        results['_n_generations'] = self._n_generations  # number of generations
        results['_list_generations'] = self._list_generations  # list of number of generations for each restart
        return results



import numpy as np

from pypop7.optimizers.ds.ds import DS


class CS(DS):
    """Coordinate Search (CS).

    .. note:: `CS` is the *earliest* Direct (Pattern) Search method, at least dating back to Fermi (`The Nobel
       Prize in Physics 1938 <https://www.nobelprize.org/prizes/physics/1938/summary/>`_) and Metropolis (`IEEE Computer
       Society Computer Pioneer Award 1984 <https://en.wikipedia.org/wiki/Computer_Pioneer_Award>`_). Given that now
       it is *rarely* used to optimize black-box problems, it is **highly recommended** to first attempt other more
       advanced methods for large-scale black-box optimization (LSBBO).

       Its original version needs `3**n - 1` samples for each iteration in the worst case, where `n` is the
       dimensionality of the problem. Such a worst-case complexity limits its applicability for LSBBO severely.
       Instead, here we use the **opportunistic** strategy for simplicity. See Algorithm 3 from `Torczon, 1997, SIOPT
       <https://epubs.siam.org/doi/abs/10.1137/S1052623493250780>`_ for more details.

       AKA alternating directions, alternating variable search, axial relaxation, local variation, compass search.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma' - initial global step-size (`float`, default: `1.0`),
                * 'x'     - initial (starting) point (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'gamma' - decreasing factor of global step-size (`float`, default: `0.5`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.ds.cs import CS
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'x': 3*numpy.ones((2,)),
       ...            'sigma': 1.0,
       ...            'verbose_frequency': 500}
       >>> cs = CS(problem, options)  # initialize the optimizer class
       >>> results = cs.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"CS: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       CS: 5000, 0.1491367032979898

    Attributes
    ----------
    gamma : `float`
            decreasing factor of global step-size.
    sigma : `float`
            final global step-size (changed during optimization).
    x     : `array_like`
            initial (starting) point.

    References
    ----------
    Larson, J., Menickelly, M. and Wild, S.M., 2019.
    Derivative-free optimization methods.
    Acta Numerica, 28, pp.287-404.
    https://tinyurl.com/4sr2t63j

    Audet, C. and Hare, W., 2017.
    Derivative-free and blackbox optimization.
    Berlin: Springer International Publishing.
    https://link.springer.com/book/10.1007/978-3-319-68913-5

    Torczon, V., 1997.
    On the convergence of pattern search algorithms.
    SIAM Journal on Optimization, 7(1), pp.1-25.
    https://epubs.siam.org/doi/abs/10.1137/S1052623493250780
    (See Algorithm 3 (Section 4.1) for details.)

    Fermi, E. and Metropolis N., 1952.
    Numerical solution of a minimum problem.
    Los Alamos Scientific Lab., Los Alamos, NM.
    https://www.osti.gov/servlets/purl/4377177
    """
    def __init__(self, problem, options):
        DS.__init__(self, problem, options)
        self.gamma = options.get('gamma', 0.5)  # decreasing factor of global step-size
        assert self.gamma > 0.0

    def initialize(self, args=None, is_restart=False):
        x = self._initialize_x(is_restart)  # initial point
        y = self._evaluate_fitness(x, args)  # fitness
        return x, y

    def iterate(self, x=None, args=None):
        improved, fitness = False, []
        for i in range(self.ndim_problem):  # to search along each coordinate
            for sgn in [-1, 1]:  # for two opponent directions
                if self._check_terminations():
                    return x, fitness
                xx = np.copy(x)
                xx[i] += sgn*self.sigma
                y = self._evaluate_fitness(xx, args)
                fitness.append(y)
                if y < self.best_so_far_y:
                    x = xx  # greedy / opportunistic
                    improved = True
                    break
        if not improved:  # to decrease step-size if no improvement
            self.sigma *= self.gamma
        return x, fitness

    def restart_reinitialize(self, args=None, x=None, y=None, fitness=None):
        self._fitness_list.append(self.best_so_far_y)
        is_restart_1, is_restart_2 = self.sigma < self.sigma_threshold, False
        if len(self._fitness_list) >= self.stagnation:
            is_restart_2 = (self._fitness_list[-self.stagnation] - self._fitness_list[-1]) < self.fitness_diff
        is_restart = bool(is_restart_1) or bool(is_restart_2)
        if is_restart:
            self._print_verbose_info(fitness, y)
            self.sigma = np.copy(self._sigma_bak)
            x, y = self.initialize(args, is_restart)
            self._fitness_list = [self.best_so_far_y]
            self._n_generations = 0
            self._n_restart += 1
            if self.verbose:
                print(' ....... *** restart *** .......')
        return x, y

    def optimize(self, fitness_function=None, args=None):
        fitness = DS.optimize(self, fitness_function)
        x, y = self.initialize(args)
        while True:
            self._print_verbose_info(fitness, y)
            x, y = self.iterate(x, args)
            if self._check_terminations():
                break
            self._n_generations += 1
            if self.is_restart:
                x, y = self.restart_reinitialize(args, x, y, fitness)
        return self._collect(fitness, y)



"""Repeat the following paper for `MMES`:
    He, X., Zheng, Z. and Zhou, Y., 2021.
    MMES: Mixture model-based evolution strategy for large-scale optimization.
    IEEE Transactions on Evolutionary Computation, 25(2), pp.320-333.
    https://ieeexplore.ieee.org/abstract/document/9244595

    All generated figures can be accessed via the following link:
    https://github.com/Evolutionary-Intelligence/pypop/tree/main/docs/repeatability/mmes

    Luckily our Python code could repeat the data reported in the paper *well*.
    Therefore, we argue that its repeatability could be **well-documented**.
"""
import pickle

import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.benchmarks.base_functions import discus, cigar, ellipsoid, rosenbrock
from pypop7.optimizers.es.mmes import MMES as Solver


def plot(function, ndim):
    plt.figure()
    result = pickle.load(open(function + '_' + str(ndim) + '.pickle', 'rb'))
    plt.yscale('log')
    plt.xscale('log')
    plt.plot(result['fitness'][:, 0], result['fitness'][:, 1], color='r')
    if function == 'discus':
        plt.xlim([1e2, 1e8])
        plt.ylim([1e-8, 1e7])
        plt.xticks([1e2, 1e3, 1e4, 1e5, 1e6, 1e7])
        plt.yticks(ticks=[1e-5, 1e0, 1e5])
    elif function == 'cigar':
        plt.xlim([1e2, 1e7])
        plt.ylim([1e-8, 1e10])
        plt.xticks([1e2, 1e3, 1e4, 1e5, 1e6])
        plt.yticks(ticks=[1e-5, 1e0, 1e5])
    elif function == 'ellipsoid':
        plt.xlim([1e2, 1e8])
        plt.ylim([1e-8, 1e9])
        plt.xticks([1e2, 1e3, 1e4, 1e5, 1e6, 1e7, 1e8])
        plt.yticks(ticks=[1e-5, 1e0, 1e5])
    elif function == 'rosenbrock':
        plt.xlim([1e2, 1e8])
        plt.ylim([1e-8, 1e8])
        plt.xticks([1e2, 1e3, 1e4, 1e5, 1e6, 1e7, 1e8])
        plt.yticks(ticks=[1e-5, 1e0, 1e5])
    plt.xlabel("FEs")
    plt.ylabel("Objective Value")
    plt.title(function.capitalize())
    plt.show()


if __name__ == '__main__':
    sns.set_theme(style='darkgrid')
    ndim_problem = 1000
    for f in [discus, cigar, ellipsoid, rosenbrock]:
        problem = {'fitness_function': f,
                   'ndim_problem': ndim_problem,
                   'lower_boundary': -5*np.ones((ndim_problem,)),
                   'upper_boundary': 5*np.ones((ndim_problem,))}
        options = {'fitness_threshold': 1e-8,
                   'max_function_evaluations': 1e8,
                   'seed_rng': 2022,  # not given in the original paper
                   'sigma': 3,
                   'saving_fitness': 1,
                   'is_restart': False}
        solver = Solver(problem, options)
        results = solver.optimize()
        pickle.dump(results, open(f.__name__ + '_' + str(ndim_problem) + '.pickle', 'wb'))
        plot(f.__name__, ndim_problem)



"""Repeat the following paper for `CSAES`:
    Ostermeier, A., Gawelczyk, A. and Hansen, N., 1994, October.
    Step-size adaptation based on non-local use of selection information
    In International Conference on Parallel Problem Solving from Nature (pp. 189-198).
    Springer, Berlin, Heidelberg.
    http://link.springer.com/chapter/10.1007/3-540-58484-6_263

    All generated figures can be accessed via the following link:
    https://github.com/Evolutionary-Intelligence/pypop/blob/main/docs/repeatability/csaes/_repeat_csaes.png

    Luckily our Python code could repeat the data reported in the original paper *well*.
    Therefore, we argue that its repeatability could be **well-documented**.
"""
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.benchmarks.base_functions import schwefel12
from pypop7.optimizers.es.csaes import CSAES


if __name__ == '__main__':
    sns.set_theme(style='darkgrid')
    plt.figure()
    problem = {'fitness_function': schwefel12,
               'ndim_problem': 20,
               'lower_boundary': -65*np.ones((20,)),
               'upper_boundary': 65*np.ones((20,))}
    options = {'max_function_evaluations': 60000,
               'fitness_threshold': 1e-3,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 2.0,  # undefined in the original paper
               'n_individuals': 10,
               'saving_fitness': 1,
               'is_restart': False}
    csaes = CSAES(problem, options)
    fitness = csaes.optimize()['fitness']
    plt.plot(fitness[:, 0], fitness[:, 1], 'k')
    plt.xticks([0, 10000, 20000, 30000, 40000, 50000, 60000])
    plt.xlim([0, 60000])
    plt.xlabel('function evaluations')
    plt.yticks([1e-3, 1e-2, 1e-1, 1e0, 1e1, 1e2, 1e3, 1e4, 1e5])
    plt.ylim([1e-3, 1e5])
    plt.yscale('log')
    plt.ylabel('best function value')
    plt.show()



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer
from pypop7.optimizers.sa.sa import SA


class CSA(SA):
    """Corana et al.' Simulated Annealing (CSA).

    .. note:: `"The algorithm is essentially an iterative random search procedure with adaptive moves along
       the coordinate directions."---[Corana et al., 1987, ACM-TOMS] <https://dl.acm.org/doi/abs/10.1145/29380.29864>`_

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'       - initial global step-size (`float`),
                * 'temperature' - annealing temperature (`float`),
                * 'n_sv'        - frequency of step variation (`int`, default: `20`),
                * 'c'           - factor of step variation (`float`, default: `2.0`),
                * 'n_tr'        - frequency of temperature reduction (`int`, default:
                                  `np.maximum(100, 5*problem['ndim_problem'])`),
                * 'f_tr'        - factor of temperature reduction (`int`, default: `0.85`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.sa.csa import CSA
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'x': 3*numpy.ones((2,)),
       ...            'sigma': 1.0,
       ...            'temperature': 100}
       >>> csa = CSA(problem, options)  # initialize the optimizer class
       >>> results = csa.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"CSA: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       CSA: 5000, 0.0023146719686626344

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/bdd62drw>`_ for more details.

    Attributes
    ----------
    c           : `float`
                  factor of step variation.
    f_tr        : `int`
                  factor of temperature reduction.
    n_sv        : `int`
                  frequency of step variation
    n_tr        : `int`
                  frequency of temperature reduction
    sigma       : `float`
                  initial global step-size.
    temperature : `float`
                  annealing temperature.

    References
    ----------
    Corana, A., Marchesi, M., Martini, C. and Ridella, S., 1987.
    Minimizing multimodal functions of continuous variables with the "simulated annealing" algorithm.
    ACM Transactions on Mathematical Software, 13(3), pp.262-280.
    https://dl.acm.org/doi/abs/10.1145/29380.29864
    https://dl.acm.org/doi/10.1145/66888.356281

    Kirkpatrick, S., Gelatt, C.D. and Vecchi, M.P., 1983.
    Optimization by simulated annealing.
    Science, 220(4598), pp.671-680.
    https://science.sciencemag.org/content/220/4598/671
    """
    def __init__(self, problem, options):
        SA.__init__(self, problem, options)
        self.sigma = options.get('sigma')
        assert self.sigma > 0.0
        self.v = self.sigma*np.ones((self.ndim_problem,))  # step vector
        self.n_sv = options.get('n_sv', 20)  # frequency of step variation (N_S)
        assert self.n_sv > 0
        self.c = options.get('c', 2.0)  # factor of step variation
        assert self.c > 0.0
        self.f_sv = self.c*np.ones(self.ndim_problem,)
        # set frequency of temperature reduction (N_T)
        self.n_tr = options.get('n_tr', np.maximum(100, 5*self.ndim_problem))
        assert self.n_tr > 0
        self.f_tr = options.get('r_T', 0.85)  # factor of temperature reduction (r_T)
        assert self.f_tr > 0.0
        self._sv = np.zeros((self.ndim_problem,))  # for step variation
        self.verbose = options.get('verbose', 10)

    def initialize(self, args=None):
        if self.x is None:  # starting point
            x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary)
        else:
            x = np.copy(self.x)
        assert len(x) == self.ndim_problem
        y = self._evaluate_fitness(x, args)
        self.parent_x, self.parent_y = np.copy(x), np.copy(y)
        return y

    def iterate(self, args=None):  # to perform a cycle of random moves
        fitness = []
        for h in np.arange(self.ndim_problem):
            if self._check_terminations():
                break
            x = np.copy(self.parent_x)
            search_range = (np.maximum(self.parent_x[h] - self.v[h], self.lower_boundary[h]),
                            np.minimum(self.parent_x[h] + self.v[h], self.upper_boundary[h]))
            x[h] = self.rng_optimization.uniform(search_range[0], search_range[1])
            y = self._evaluate_fitness(x, args)
            fitness.append(y)
            diff = self.parent_y - y
            if (diff >= 0) or (self.rng_optimization.random() < np.exp(diff/self.temperature)):
                self.parent_x, self.parent_y = np.copy(x), np.copy(y)
                self._sv[h] += 1
        return fitness

    def _adjust_step_vector(self):
        for u in range(self.ndim_problem):
            if self._sv[u] > 0.6*self.n_sv:
                self.v[u] *= 1.0 + self.f_sv[u]*(self._sv[u]/self.n_sv - 0.6)/0.4
            elif self._sv[u] < 0.4*self.n_sv:
                self.v[u] /= 1.0 + self.f_sv[u]*(0.4 - self._sv[u]/self.n_sv)/0.4
        self._sv = np.zeros((self.ndim_problem,))

    def optimize(self, fitness_function=None, args=None):  # for all iterations (generations)
        fitness = Optimizer.optimize(self, fitness_function)
        y = self.initialize(args)
        self._print_verbose_info(fitness, y)
        while not self._check_terminations():
            for m in range(self.n_tr):
                if self._check_terminations():
                    break
                for j in range(self.n_sv):
                    if self._check_terminations():
                        break
                    y = self.iterate(args)
                    self._n_generations += 1
                    self._print_verbose_info(fitness, y)
                self._adjust_step_vector()
            self.temperature *= self.f_tr  # temperature reducing
            self.parent_x, self.parent_y = np.copy(self.best_so_far_x), np.copy(self.best_so_far_y)
        results = self._collect(fitness)
        results['v'] = np.copy(self.v)
        return results



"""This is a simple demo for `PyPop7` optimization on the well-designed `COCO` platform.

    If you cannot install `cocoex` on Windows, the solution is given in:
        https://github.com/numbbo/coco/issues/2086

    Note that this script is only used to check whether or not `cocoex` is installed
    *successfully*. For benchmarking on `cocoex`, please run the following code:
    https://github.com/Evolutionary-Intelligence/pypop/blob/main/tutorials/coco_benchmarking.py

    References
    ----------
    N. Hansen, A. Auger, R. Ros, O. Mersmann, T. Tušar, D. Brockhoff.
    COCO: A Platform for Comparing Continuous Optimizers in a Black-Box Setting.
    Optimization Methods and Software, 36(1), pp.114-144, 2021.
"""
import cocoex
import numpy as np

from pypop7.optimizers.ds.nm import NM as Solver


if __name__ == '__main__':
    print(cocoex.known_suite_names)
    suite = cocoex.Suite('bbob', '', '')
    for current_problem in suite:
        print(current_problem)
        d = current_problem.dimension
        problem = {'fitness_function': current_problem,
                   'ndim_problem': d,
                   'lower_boundary': -10*np.ones((d,)),
                   'upper_boundary': 10*np.ones((d,))}
        options = {'max_function_evaluations': 1e3*d,
                   'seed_rng': 2022,
                   'sigma': 1.0,
                   'verbose': False,
                   'saving_fitness': 2000}
        solver = Solver(problem, options)
        results = solver.optimize()
        print('  best-so-far fitness:', results['best_so_far_y'])



from pypop7.optimizers.core.optimizer import Optimizer


class BO(Optimizer):
    """Bayesian Optimization (BO).

    References
    ----------
    https://bayesoptbook.com/

    https://bayesopt-tutorial.github.io/
    """
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)

    def initialize(self):
        raise NotImplementedError

    def iterate(self):  # for each iteration (generation)
        raise NotImplementedError



import numpy as np

from pypop7.optimizers.es.es import ES


class DDCMA(ES):
    """Diagonal Decoding Covariance Matrix Adaptation (DDCMA).

    .. note:: `DDCMA` is a *state-of-the-art* improvement version of the well-designed `CMA-ES` algorithm, which enjoys
       both two worlds of `SEP-CMA-ES` (faster adaptation on nearly separable problems) and `CMA-ES` (more robust
       adaptation on ill-conditioned non-separable problems) via **adaptive diagonal decoding**. It is **highly
       recommended** to first attempt other ES variants (e.g., `LMCMA`, `LMMAES`) for large-scale black-box
       optimization, since `DDCMA` has a *quadratic* time complexity (w.r.t. each sampling).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'         - initial global step-size, aka mutation strength (`float`),
                * 'mean'          - initial (starting) point, aka mean of Gaussian search distribution (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'n_individuals' - number of offspring, aka offspring population size (`int`, default:
                  `4 + int(3*np.log(problem['ndim_problem']))`).

    Examples
    --------
    Use the optimizer `DDCMA` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.ddcma import DDCMA
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'is_restart': False,
       ...            'mean': 3*numpy.ones((2,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> ddcma = DDCMA(problem, options)  # initialize the optimizer class
       >>> results = ddcma.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"DDCMA: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       DDCMA: 5000, 0.0

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/mc34kkmn>`_ for more details.

    Attributes
    ----------
    mean          : `array_like`
                    initial (starting) point, aka mean of Gaussian search distribution.
    n_individuals : `int`
                    number of offspring, aka offspring population size.
    sigma         : `float`
                    final global step-size, aka mutation strength.

    References
    ----------
    Akimoto, Y. and Hansen, N., 2020.
    Diagonal acceleration for covariance matrix adaptation evolution strategies.
    Evolutionary Computation, 28(3), pp.405-435.
    https://direct.mit.edu/evco/article/28/3/405/94999/Diagonal-Acceleration-for-Covariance-Matrix

    See its official Python implementation from Prof. Akimoto:
    https://gist.github.com/youheiakimoto/1180b67b5a0b1265c204cba991fa8518
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self._mu_eff, self._mu_eff_negative = None, None
        self.c_s, self.d_s = None, None
        self._gamma_s, self._gamma_c, self._gamma_d = 0.0, 0.0, 0.0
        self.c_1, self.c_w, self.c_c = None, None, None
        self._w, self._w_d = None, None
        self.c_1_d, self.c_w_d, self.c_c_d = None, None, None
        self._beta_eig, self._t_eig = None, None
        self._n_generations = 0

    def _set_c_1_and_c_1_d(self, m):
        return 1.0/(2.0*(m/self.ndim_problem + 1.0)*np.power((self.ndim_problem + 1.0), 0.75) + self._mu_eff/2.0)

    def initialize(self, is_restart=False):
        w_apostrophe = np.log((self.n_individuals + 1.0)/2.0) - np.log(np.arange(self.n_individuals) + 1.0)
        positive_w, negative_w = w_apostrophe > 0, w_apostrophe < 0
        w_apostrophe[positive_w] /= np.sum(np.abs(w_apostrophe[positive_w]))
        w_apostrophe[negative_w] /= np.sum(np.abs(w_apostrophe[negative_w]))
        self._mu_eff = 1.0/np.sum(np.power(w_apostrophe[positive_w], 2))
        self._mu_eff_negative = 1.0/np.sum(np.power(w_apostrophe[negative_w], 2))
        self.c_s = (self._mu_eff + 2.0)/(self.ndim_problem + self._mu_eff + 5.0)
        self.d_s = 1.0 + self.c_s + 2.0*np.maximum(0.0, np.sqrt((self._mu_eff - 1.0)/(self.ndim_problem + 1.0)) - 1.0)
        mu_apostrophe = self._mu_eff + 1.0/self._mu_eff - 2.0 + self.n_individuals/(2.0*(self.n_individuals + 5.0))
        m = self.ndim_problem*(self.ndim_problem + 1)/2
        self._gamma_s, self._gamma_c, self._gamma_d = 0.0, 0.0, 0.0
        self.c_1 = self._set_c_1_and_c_1_d(m)
        self.c_w = np.minimum(mu_apostrophe*self.c_1, 1.0 - self.c_1)
        self.c_c = np.sqrt(self._mu_eff*self.c_1)/2.0
        self._w = np.copy(w_apostrophe)
        self._w[negative_w] *= np.minimum(1.0 + self.c_1/self.c_w, 1.0 + 2.0*self._mu_eff_negative/(self._mu_eff + 2.0))
        m = self.ndim_problem
        self.c_1_d = self._set_c_1_and_c_1_d(m)
        self.c_w_d = np.minimum(mu_apostrophe*self.c_1_d, 1.0 - self.c_1_d)
        self.c_c_d = np.sqrt(self._mu_eff*self.c_1_d)/2.0
        self._w_d = np.copy(w_apostrophe)
        self._w_d[negative_w] *= np.minimum(1.0 + self.c_1_d/self.c_w_d,
                                            1.0 + 2.0*self._mu_eff_negative/(self._mu_eff + 2.0))
        self._beta_eig = 10*self.ndim_problem
        self._t_eig = np.maximum(1.0, np.floor(1.0/(self._beta_eig*(self.c_1 + self.c_w))))
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        d = self.sigma*np.ones((self.ndim_problem,))  # diagonal decoding matrix
        self.sigma = 1.0
        sqrt_c = np.eye(self.ndim_problem)
        inv_sqrt_c = np.eye(self.ndim_problem)
        z = np.empty((self.n_individuals, self.ndim_problem))
        cz = np.empty((self.n_individuals, self.ndim_problem))
        x = np.empty((self.n_individuals, self.ndim_problem))
        y = np.empty((self.n_individuals,))
        p_s = np.zeros((self.ndim_problem,))
        p_c = np.zeros((self.ndim_problem,))
        p_c_d = np.zeros((self.ndim_problem,))
        cm = np.eye(self.ndim_problem)
        sqrt_eig_va = np.ones((self.ndim_problem,))
        self._list_initial_mean.append(np.copy(mean))
        self._n_generations = 0
        return mean, d, sqrt_c, inv_sqrt_c, z, cz, x, y, p_s, p_c, p_c_d, cm, sqrt_eig_va

    def iterate(self, mean=None, d=None, sqrt_c=None, z=None, cz=None, x=None, y=None, args=None):
        for k in range(self.n_individuals):  # to sample offspring population
            if self._check_terminations():
                return z, cz, x, y
            z[k] = self.rng_optimization.standard_normal((self.ndim_problem,))  # Gaussian noise for mutation
            cz[k] = np.dot(z[k], sqrt_c)
            x[k] = mean + self.sigma*d*cz[k]  # offspring individual
            y[k] = self._evaluate_fitness(x[k], args)  # fitness
        self._n_generations += 1
        return z, cz, x, y

    def _update_distribution(self, mean=None, d=None, sqrt_c=None, inv_sqrt_c=None, z=None, cz=None,
                             x=None, y=None, p_s=None, p_c=None, p_c_d=None, cm=None, sqrt_eig_va=None):
        order = np.argsort(y)
        zz, xx = z[order], x[order]
        positive_w = self._w > 0
        wz = np.dot(self._w[positive_w], zz[positive_w])
        wcz = np.dot(self._w[positive_w], cz[order][positive_w])
        # update distribution mean via weighted multi-recombination
        mean += self.sigma*d*wcz
        # update global step-size via CSA
        p_s = (1.0 - self.c_s)*p_s + np.sqrt(self.c_s*(2.0 - self.c_s)*self._mu_eff)*wz
        self._gamma_s = np.power(1.0 - self.c_s, 2)*self._gamma_s + self.c_s*(2.0 - self.c_s)
        self.sigma *= np.exp(self.c_s/self.d_s*(np.linalg.norm(p_s)/self._e_chi - np.sqrt(self._gamma_s)))
        # update evolution path
        h_s = np.dot(p_s, p_s)/self._gamma_s < (2.0 + 4.0/(self.ndim_problem + 1.0))*self.ndim_problem
        p_c = (1.0 - self.c_c)*p_c + h_s*np.sqrt(self.c_c*(2.0 - self.c_c)*self._mu_eff)*d*wcz
        self._gamma_c = np.power(1.0 - self.c_c, 2)*self._gamma_c + h_s*self.c_c*(2.0 - self.c_c)
        # update covariance matrix
        pw, nw = self._w > 0, self._w < 0
        d_p = np.dot(p_c/d, inv_sqrt_c)
        cm += self.c_1*(np.outer(d_p, d_p) - self._gamma_c*np.eye(self.ndim_problem)) + self.c_w*(
                np.dot(np.transpose(zz[pw])*self._w[pw], zz[pw]) - np.sum(self._w[pw])*np.eye(self.ndim_problem) +
                np.dot(np.transpose(zz[nw])*(self._w[nw]*self.ndim_problem/np.power(np.linalg.norm(zz[nw], axis=1), 2)),
                       zz[nw]) - np.sum(self._w[nw])*np.eye(self.ndim_problem))  # Eq.19
        p_c_d = (1.0 - self.c_c_d)*p_c_d + h_s*np.sqrt(self.c_c_d*(2 - self.c_c_d)*self._mu_eff)*d*wcz
        self._gamma_d = np.power(1 - self.c_c_d, 2)*self._gamma_d + h_s*self.c_c_d*(2.0 - self.c_c_d)
        pwd, nwd = self._w_d > 0, self._w_d < 0
        eig_va = self.c_1_d*(np.power(np.dot(p_c_d/d, inv_sqrt_c), 2) - self._gamma_d) + self.c_w_d*(
                np.dot(self._w_d[pwd], np.power(zz[pwd], 2)) + (np.dot(self._w_d[nwd]*self.ndim_problem/np.power(
                    np.linalg.norm(zz[nwd], axis=1), 2), np.power(zz[nwd], 2)) - np.sum(self._w_d)))  # Eq.28
        d *= np.exp(eig_va/(2.0*np.maximum(1.0, np.max(sqrt_eig_va)/np.min(sqrt_eig_va) - 2.0 + 1.0)))  # Eq.29 + Eq.31
        if self._n_generations % self._t_eig == 0:
            c = np.dot(np.dot(sqrt_c, np.eye(self.ndim_problem) + np.minimum(
                0.75/np.abs(np.min(np.linalg.eigvalsh(cm))), 1.0)*cm), sqrt_c)  # Eq.21 + Eq.22
            sqrt_diag = np.sqrt(np.diag(c))
            d *= sqrt_diag  # Eq.34
            c = np.transpose(c/sqrt_diag)/sqrt_diag  # Eq.35 (correlation matrix)
            eig_va, eig_ve = np.linalg.eigh(c)  # to perform eigen decomposition
            sqrt_eig_va = np.sqrt(eig_va)
            sqrt_c = np.dot(eig_ve*sqrt_eig_va, np.transpose(eig_ve))
            inv_sqrt_c = np.dot(eig_ve/sqrt_eig_va, np.transpose(eig_ve))
            cm[:, :] = 0.0
        return mean, d, sqrt_c, inv_sqrt_c, cz, p_s, p_c, p_c_d, cm, sqrt_eig_va

    def restart_reinitialize(self, mean=None, d=None, sqrt_c=None, inv_sqrt_c=None, z=None, cz=None,
                             x=None, y=None, p_s=None, p_c=None, p_c_d=None, cm=None, sqrt_eig_va=None):
        if self.is_restart and ES.restart_reinitialize(self, y):
            mean, d, sqrt_c, inv_sqrt_c, z, cz, x, y, p_s, p_c, p_c_d, cm, sqrt_eig_va = self.initialize(True)
        return mean, d, sqrt_c, inv_sqrt_c, z, cz, x, y, p_s, p_c, p_c_d, cm, sqrt_eig_va

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        mean, d, sqrt_c, inv_sqrt_c, z, cz, x, y, p_s, p_c, p_c_d, cm, sqrt_eig_va = self.initialize()
        while not self.termination_signal:
            # sample and evaluate offspring population
            z, cz, x, y = self.iterate(mean, d, sqrt_c, z, cz, x, y, args)
            if self._check_terminations():
                break
            mean, d, sqrt_c, inv_sqrt_c, cz, p_s, p_c, p_c_d, cm, sqrt_eig_va = self._update_distribution(
                mean, d, sqrt_c, inv_sqrt_c, z, cz, x, y, p_s, p_c, p_c_d, cm, sqrt_eig_va)
            self._print_verbose_info(fitness, y)
            mean, d, sqrt_c, inv_sqrt_c, z, cz, x, y, p_s, p_c, p_c_d, cm, sqrt_eig_va = self.restart_reinitialize(
                mean, d, sqrt_c, inv_sqrt_c, z, cz, x, y, p_s, p_c, p_c_d, cm, sqrt_eig_va)
        results = self._collect(fitness, y, mean)
        # by default, do NOT save covariance matrix of search distribution in order to save memory,
        # owing to its *quadratic* space complexity
        return results



"""Repeat the following paper for `G3PCX`:
    Deb, K., Anand, A. and Joshi, D., 2002.
    A computationally efficient evolutionary algorithm for real-parameter optimization.
    Evolutionary Computation, 10(4), pp.371-395.
    https://direct.mit.edu/evco/article-abstract/10/4/371/1136/A-Computationally-Efficient-Evolutionary-Algorithm
    https://www.egr.msu.edu/~kdeb/codes/g3pcx/g3pcx.tar  (See the original C source code.)

    All generated figures can be accessed via the following link:
    https://github.com/Evolutionary-Intelligence/pypop/blob/main/docs/repeatability/g3pcx/_repeat_g3pcx.png

    Luckily our Python code could repeat the data reported in the original paper *well*.
    Therefore, we argue that its repeatability could be **well-documented**.
"""
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.benchmarks.base_functions import schwefel12
from pypop7.optimizers.ga.g3pcx import G3PCX as Solver


if __name__ == '__main__':
    ndim_problem = 20
    for f in [schwefel12]:
        problem = {'fitness_function': f,
                   'ndim_problem': ndim_problem,
                   'initial_lower_boundary': -10*np.ones((ndim_problem,)),
                   'initial_upper_boundary': -5*np.ones((ndim_problem,))}
        options = {'max_function_evaluations': 50000,
                   'fitness_threshold': 1e-20,
                   'seed_rng': 0,
                   'saving_fitness': 1}
        solver = Solver(problem, options)
        results = solver.optimize()

        sns.set_theme(style='darkgrid')
        plt.figure()
        fitness = results['fitness']
        plt.plot(fitness[:, 0], fitness[:, 1], 'k')
        plt.xticks([0, 50000, 100000, 150000, 200000])
        plt.xlabel('Function Evaluations')
        plt.yscale('log')
        plt.yticks([1e-20, 1e-15, 1e-10, 1e-5, 1e0, 1e5])
        plt.ylabel('Best Fitness')
        plt.show()



import numpy as np

from pypop7.optimizers.es.es import ES


class MAES(ES):
    """Matrix Adaptation Evolution Strategy (MAES).

    .. note:: `MAES` is a powerful *simplified* version of the well-established `CMA-ES` algorithm nearly without
       significant performance loss, designed in 2017 by `Beyer <https://homepages.fhv.at/hgb/>`_ and `Sendhoff
       <https://tinyurl.com/26szwuaa>`_. One obvious advantage of such a simplification is to help better understand
       the underlying working principles (e.g., **invariance** and **unbias**) of `CMA-ES`, which are often thought to
       be rather complex for newcomers. It is **highly recommended** to first attempt more advanced ES variants (e.g.,
       `LMCMA`, `LMMAES`) for large-scale black-box optimization, since `MAES` has a *cubic* time complexity (w.r.t.
       each sampling). Note that another improved version called `FMAES` provides a *relatively more efficient*
       implementation for `MAES` with *quadratic* time complexity (w.r.t. each sampling).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'         - initial global step-size, aka mutation strength (`float`),
                * 'mean'          - initial (starting) point, aka mean of Gaussian search distribution (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'n_individuals' - number of offspring, aka offspring population size (`int`, default:
                  `4 + int(3*np.log(problem['ndim_problem']))`),
                * 'n_parents'     - number of parents, aka parental population size (`int`, default:
                  `int(options['n_individuals']/2)`).

    Examples
    --------
    Use the optimizer `MAES` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.maes import MAES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3*numpy.ones((2,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> maes = MAES(problem, options)  # initialize the optimizer class
       >>> results = maes.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"MAES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       MAES: 5000, 2.129367016460251e-19

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/3zvve79b>`_ for more details.

    Attributes
    ----------
    mean          : `array_like`
                    initial (starting) point, aka mean of Gaussian search distribution.
    n_individuals : `int`
                    number of offspring, aka offspring population size.
    n_parents     : `int`
                    number of parents, aka parental population size.
    sigma         : `float`
                    final global step-size, aka mutation strength.

    References
    ----------
    Beyer, H.G., 2020, July.
    Design principles for matrix adaptation evolution strategies.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation Companion (pp. 682-700).
    https://dl.acm.org/doi/abs/10.1145/3377929.3389870

    Loshchilov, I., Glasmachers, T. and Beyer, H.G., 2019.
    Large scale black-box optimization by limited-memory matrix adaptation.
    IEEE Transactions on Evolutionary Computation, 23(2), pp.353-358.
    https://ieeexplore.ieee.org/abstract/document/8410043

    Beyer, H.G. and Sendhoff, B., 2017.
    Simplify your covariance matrix adaptation evolution strategy.
    IEEE Transactions on Evolutionary Computation, 21(5), pp.746-759.
    https://ieeexplore.ieee.org/document/7875115

    See the official Matlab version from Prof. Beyer:
    https://homepages.fhv.at/hgb/downloads/ForDistributionFastMAES.tar
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.options = options
        self.c_s = None
        self.c_1 = None
        self.c_w = None
        self.d_sigma = None
        self._alpha_cov = 2.0
        self._s_1 = None
        self._s_2 = None
        self._fast_version = options.get('_fast_version', False)
        if not self._fast_version:
            self._diag_one = np.eye(self.ndim_problem)

    def _set_c_w(self):
        return np.minimum(1.0 - self.c_1, self._alpha_cov*(self._mu_eff + 1.0/self._mu_eff - 2.0) /
                          (np.power(self.ndim_problem + 2.0, 2) + self._alpha_cov*self._mu_eff/2.0))

    def _set_d_sigma(self):
        return 1.0 + self.c_s + 2.0*np.maximum(0.0, np.sqrt((self._mu_eff - 1.0)/(self.ndim_problem + 1.0)) - 1.0)

    def initialize(self, is_restart=False):
        self.c_s = self.options.get('c_s', (self._mu_eff + 2.0)/(self._mu_eff + self.ndim_problem + 5.0))
        self.c_1 = self.options.get('c_1', self._alpha_cov/(np.power(self.ndim_problem + 1.3, 2) + self._mu_eff))
        self.c_w = self.options.get('c_w', self._set_c_w())
        self.d_sigma = self.options.get('d_sigma', self._set_d_sigma())
        self._s_1 = 1.0 - self.c_s
        self._s_2 = np.sqrt(self._mu_eff*self.c_s*(2.0 - self.c_s))
        z = np.empty((self.n_individuals, self.ndim_problem))  # Gaussian noise for mutation
        d = np.empty((self.n_individuals, self.ndim_problem))  # search directions
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        s = np.zeros((self.ndim_problem,))  # evolution path
        tm = np.eye(self.ndim_problem)  # transformation matrix
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        self._list_initial_mean.append(np.copy(mean))
        return z, d, mean, s, tm, y

    def iterate(self, z=None, d=None, mean=None, tm=None, y=None, args=None):
        for k in range(self.n_individuals):  # to sample offspring population
            if self._check_terminations():
                return z, d, y
            z[k] = self.rng_optimization.standard_normal((self.ndim_problem,))
            d[k] = np.dot(tm, z[k])
            y[k] = self._evaluate_fitness(mean + self.sigma*d[k], args)
        return z, d, y

    def _update_distribution(self, z=None, d=None, mean=None, s=None, tm=None, y=None):
        order = np.argsort(y)
        d_w, z_w, zz_w = np.zeros((self.ndim_problem,)), np.zeros((self.ndim_problem,)), None
        if not self._fast_version:
            zz_w = np.zeros((self.ndim_problem, self.ndim_problem))
        for k in range(self.n_parents):
            d_w += self._w[k]*d[order[k]]
            z_w += self._w[k]*z[order[k]]
            if not self._fast_version:
                zz_w += self._w[k]*np.outer(z[order[k]], z[order[k]])
        # update distribution mean
        mean += self.sigma*d_w
        # update evolution path (s) and transformation matrix (M)
        s = self._s_1*s + self._s_2*z_w
        if not self._fast_version:
            tm_1 = self.c_1*(np.outer(s, s) - self._diag_one)
            tm_2 = self.c_w*(zz_w - self._diag_one)
            tm += 0.5*np.dot(tm, tm_1 + tm_2)
        else:
            tm = (1.0 - 0.5*(self.c_1 + self.c_w))*tm + (0.5*self.c_1)*np.dot(
                np.dot(tm, s[:, np.newaxis]), s[np.newaxis, :])
            for k in range(self.n_parents):
                tm += (0.5*self.c_w)*self._w[k]*np.outer(d[order[k]], z[order[k]])
        # update global step-size
        self.sigma *= np.exp(self.c_s/self.d_sigma*(np.linalg.norm(s)/self._e_chi - 1.0))
        return mean, s, tm

    def restart_reinitialize(self, z=None, d=None, mean=None, s=None, tm=None, y=None):
        if ES.restart_reinitialize(self, y):
            z, d, mean, s, tm, y = self.initialize(True)
        return z, d, mean, s, tm, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        z, d, mean, s, tm, y = self.initialize()
        while not self.termination_signal:
            # sample and evaluate offspring population
            z, d, y = self.iterate(z, d, mean, tm, y, args)
            if self._check_terminations():
                break
            mean, s, tm = self._update_distribution(z, d, mean, s, tm, y)
            self._print_verbose_info(fitness, y)
            self._n_generations += 1
            if self.is_restart:
                z, d, mean, s, tm, y = self.restart_reinitialize(z, d, mean, s, tm, y)
        results = self._collect(fitness, y, mean)
        results['s'] = s
        # by default, do NOT save transformation matrix of search distribution in order to save memory,
        # owing to its *quadratic* space complexity
        return results



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer
from pypop7.optimizers.sa.sa import SA


class NSA(SA):
    """Noisy Simulated Annealing (NSA).

    .. note:: This is a *slightly modified* version of discrete `NSA` for continuous optimization.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'x'         - initial (starting) point (`array_like`),
                * 'sigma'     - initial global step-size (`float`),
                * 'is_noisy'  - whether or not to minimize a **noisy** cost function (`bool`, default: `False`),
                * 'schedule'  - schedule for sampling intensity (`str`, default: `linear`),

                  * currently only two (*linear* or *quadratic*) schedules are supported for sampling intensity,

                * 'n_samples' - number of samples (`int`),
                * 'rt'        - reducing factor of annealing temperature (`float`, default: `0.99`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.sa.nsa import NSA
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'x': 3*numpy.ones((2,)),
       ...            'sigma': 1.0,
       ...            'temperature': 100.0}
       >>> nsa = NSA(problem, options)  # initialize the optimizer class
       >>> results = nsa.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"NSA: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       NSA: 5000, 0.006086567926462302

    For its correctness checking of coding, the *code-based repeatability report* cannot be provided owing to
    the lack of some details of its experiments in the original paper.

    Attributes
    ----------
    is_noisy  : `bool`
                whether or not to minimize a noisy cost function.
    n_samples : `int`
                number of samples for each iteration.
    rt        : `float`
                reducing factor of annealing temperature.
    schedule  : `str`
                schedule for sampling intensity.
    sigma     : `float`
                global step-size (fixed during optimization).
    x         : `array_like`
                initial (starting) point.

    References
    ----------
    Bouttier, C. and Gavra, I., 2019.
    Convergence rate of a simulated annealing algorithm with noisy observations.
    Journal of Machine Learning Research, 20(1), pp.127-171.
    https://www.jmlr.org/papers/v20/16-588.html
    """
    def __init__(self, problem, options):
        SA.__init__(self, problem, options)
        self.sigma = options.get('sigma')
        assert self.sigma > 0.0
        self.is_noisy = options.get('is_noisy', False)
        assert self.is_noisy in [True, False]
        self.schedule = options.get('schedule', 'linear')  # schedule for sampling intensity
        assert self.schedule in ['linear', 'quadratic'],\
            'Currently only two (*linear* or *quadratic*) schedules are supported for sampling intensity.'
        if self.is_noisy:
            self.n_samples = options.get('n_samples')
        else:
            self.n_samples = 1  # a mandatory setting
        assert self.n_samples > 0
        self.rt = options.get('cr', 0.99)  # reducing factor of temperature
        self._tk = 0

    def initialize(self, args=None):
        if self.x is None:  # starting point
            x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary)
        else:
            x = np.copy(self.x)
        assert len(x) == self.ndim_problem
        y = self._evaluate_fitness(x, args)
        self.parent_x, self.parent_y = np.copy(x), np.copy(y)
        return y

    def iterate(self, args=None):
        x = self.parent_x + self.sigma*self.rng_optimization.standard_normal((self.ndim_problem,))
        self._tk += self.rng_optimization.exponential()
        if self.schedule == 'linear':
            n_tk = self._tk
        else:  # quadratic
            n_tk = np.square(self._tk)
        if self.n_samples is None:
            n_samples = self.rng_optimization.poisson(n_tk) + 1
        else:
            n_samples = self.n_samples
        y, parent_y = [], []
        for _ in range(n_samples):
            if self._check_terminations():
                break
            y.append(self._evaluate_fitness(x, args))
        if self.is_noisy:  # for noisy optimization
            for _ in range(n_samples):
                if self._check_terminations():
                    break
                parent_y.append(self._evaluate_fitness(self.parent_x, args))
        else:  # for static optimization
            parent_y = self.parent_y
        diff = np.mean(parent_y) - np.mean(y)
        if (diff >= 0) or (self.rng_optimization.random() < np.exp(diff/self.temperature)):
            self.parent_x, self.parent_y = np.copy(x), np.copy(y)
        if not self.is_noisy:  # for static optimization
            parent_y = []
        if len(parent_y) > 0:
            y.extend(parent_y)
        return y

    def optimize(self, fitness_function=None, args=None):
        fitness = Optimizer.optimize(self, fitness_function)
        y = self.initialize(args)
        while not self._check_terminations():
            self._print_verbose_info(fitness, y)
            y = self.iterate(args)
            self._n_generations += 1
            self.temperature = np.maximum(self.temperature*self.rt, 1e-200)
        return self._collect(fitness, y)



import numpy as np
from scipy.stats import multivariate_normal as mn

from pypop7.optimizers.cem import CEM


class MRAS(CEM):
    """Model Reference Adaptive Search (MRAS).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'         - initial global step-size, aka mutation strength (`float`),
                * 'mean'          - initial (starting) point, aka mean of Gaussian search distribution (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'n_individuals' - number of offspring, aka offspring population size (`int`, default: `1000`),
                * 'p'             - percentage of samples as parents (`int`, default: `0.1`),
                * 'alpha'         - increasing factor of samples/individuals (`float`, default: `1.1`),
                * 'v'             - smoothing factor for search distribution update (`float`, default: `0.2`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.cem.mras import MRAS
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'sigma': 10}  # the global step-size may need to be tuned for better performance
       >>> mras = MRAS(problem, options)  # initialize the optimizer class
       >>> results = mras.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"MRAS: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       MRAS: 5000, 0.18363570418709932

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/yv44nbwu>`_ for more details.

    Attributes
    ----------
    alpha         : `float`
                    increasing factor of samples/individuals.
    mean          : `array_like`
                    initial (starting) point, aka mean of Gaussian search distribution.
    n_individuals : `int`
                    number of offspring, aka offspring population size.
    p             : `float`
                    percentage of samples as parents.
    sigma         : `float`
                    initial global step-size, aka mutation strength,
    v             : `float`
                    smoothing factor for search distribution update.

    References
    ----------
    Hu, J., Fu, M.C. and Marcus, S.I., 2007.
    A model reference adaptive search method for global optimization.
    Operations Research, 55(3), pp.549-568.
    https://pubsonline.informs.org/doi/abs/10.1287/opre.1060.0367
    """
    def __init__(self, problem, options):
        CEM.__init__(self, problem, options)
        self.p = options.get('p', 0.1)  # percentage of samples as parents
        assert 0.0 < self.p <= 1.0
        self.alpha = options.get('alpha', 1.1)  # increasing factor of samples/individuals
        assert self.alpha > 1.0
        self.v = options.get('v', 0.2)  # smoothing factor for search distribution update
        self.r = options.get('r', 1e-4)
        self.epsilon = options.get('epsilon', 1e-5)
        assert self.epsilon >= 0.0
        self._lambda = 0.01
        self._initial_pdf = None
        self._gamma = None
        self._min_elitists = 5*self.ndim_problem

    def initialize(self, is_restart=False):
        mean, cov = self._initialize_mean(is_restart), np.diag((self.sigma**2)*np.ones((self.ndim_problem,)))
        x = np.empty((self.n_individuals, self.ndim_problem))  # samples (population)
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        self._initial_pdf = [np.copy(mean), np.copy(cov)]
        return mean, cov, x, y

    def iterate(self, mean=None, cov=None, x=None, y=None, args=None):
        for i in range(self.n_individuals):
            if self._check_terminations():
                return mean, cov, x, y
            # use a very simple way to generate a mixed pdf
            if self.rng_optimization.uniform() <= 1.0 - self._lambda:
                x[i] = self.rng_optimization.multivariate_normal(mean=mean, cov=cov)
            else:
                x[i] = self.rng_optimization.multivariate_normal(
                    mean=self._initial_pdf[0], cov=self._initial_pdf[1])
            y[i] = self._evaluate_fitness(x[i], args)
        order = np.argsort(y)
        gamma = y[order[int(np.ceil(self.p*self.n_individuals))]]  # fitness threshold for top-p elitists
        if self._n_generations == 0 or gamma <= self._gamma + self.epsilon/2:  # Step 3(a)
            self._gamma = gamma
        else:  # Step 3(b)
            yy, _p = np.Inf, np.Inf
            # this is not detailed in the original paper (but may not be a key setting)
            for i in np.linspace(self.p, 0, 100)[1:]:
                yy = y[order[int(np.ceil(i*self.n_individuals))]]
                if yy <= self._gamma + self.epsilon/2:
                    _p = i
                    break
            if 0 < _p < self.p:  # Step 3(c)
                self._gamma, self.p = yy, _p
            else:
                self.n_individuals = int(np.ceil(self.alpha*self.n_individuals))
        if np.sum(y <= self._gamma) > self._min_elitists:
            _mean, _cov = np.zeros((self.ndim_problem,)), np.zeros((self.ndim_problem, self.ndim_problem))
            _n, weights, pdfs = 0, np.zeros((len(y),)), np.zeros((len(y),))
            for i in range(len(y)):
                if y[i] <= self._gamma:
                    _n += 1
                    pdfs[i] = ((1.0 - self._lambda)*mn.pdf(x[i], mean=mean, cov=cov) + self._lambda *
                               mn.pdf(x[i], mean=self._initial_pdf[0], cov=self._initial_pdf[1]))
                    weights[i] = np.power(np.exp(-self.r*y[i]), self._n_generations)/pdfs[i]
                    _mean += weights[i]*x[i]
            if _n > 0:
                _mean = (_mean/_n)/(np.sum(weights)/_n)
                for i in range(len(y)):
                    if y[i] <= self._gamma:
                        xm = x[i] - _mean
                        _cov += weights[i]*np.dot(xm[:, np.newaxis], xm[np.newaxis, :])
                _cov = (_cov/_n)/(np.sum(weights)/_n)
                mean = self.v*_mean + (1.0 - self.v)*mean
                cov = self.v*_cov + (1.0 - self.v)*cov
        return mean, cov, x, y

    def optimize(self, fitness_function=None, args=None):
        fitness = CEM.optimize(self, fitness_function)
        mean, cov, x, y = self.initialize()
        while True:
            mean, cov, x, y = self.iterate(mean, cov, x, y, args)
            self._print_verbose_info(fitness, y)
            if self._check_terminations():
                break
            self._n_generations += 1
            if self.n_individuals > len(y):
                x = np.empty((self.n_individuals, self.ndim_problem))  # samples (population)
                y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        return self._collect(fitness, y, mean)



"""Repeat the following paper for `EMNA`:
    Larrañaga, P. and Lozano, J.A. eds., 2001.
    Estimation of distribution algorithms: A new tool for evolutionary computation.
    Springer Science & Business Media.
    https://link.springer.com/book/10.1007/978-1-4615-1539-5
    (See Chapter 8 Experimental Results in Function Optimization with EDAs in Continuous Domain.)

    There seems to be vey small performance gaps between the original paper and our Python code, which
    may be ignored for benchmarking. Such gaps may attribute to a slight implementation difference
    (where our Python code does not use *ad-hoc elitist selection* employed in the original paper).
    We argue that its repeatability can be **well-documented**.
"""
import numpy as np

from pypop7.benchmarks.base_functions import rosenbrock, griewank
from pypop7.optimizers.eda.emna import EMNA


if __name__ == '__main__':
    ndim_problem = 10

    problem = {'fitness_function': rosenbrock,
               'ndim_problem': ndim_problem,
               'lower_boundary': -10*np.ones((ndim_problem,)),
               'upper_boundary': 10*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 301850,
               'n_individuals': 2000,
               'seed_rng': 0}  # undefined in the original paper
    emna = EMNA(problem, options)
    results = emna.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 7.674018032114779 vs 8.7201 (from the original paper)

    problem = {'fitness_function': griewank,
               'ndim_problem': ndim_problem,
               'lower_boundary': -600*np.ones((ndim_problem,)),
               'upper_boundary': 600*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 301850,
               'n_individuals': 750,
               'seed_rng': 0}  # undefined in the original paper
    emna = EMNA(problem, options)
    results = emna.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 0.0 vs 5.1166e-02 (from the original paper)

    ndim_problem = 50

    problem = {'fitness_function': rosenbrock,
               'ndim_problem': ndim_problem,
               'lower_boundary': -10*np.ones((ndim_problem,)),
               'upper_boundary': 10*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 301850,
               'n_individuals': 2000,
               'seed_rng': 0}  # undefined in the original paper
    emna = EMNA(problem, options)
    results = emna.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 56.71781067326537 vs 49.7588 (from the original paper)

    problem = {'fitness_function': griewank,
               'ndim_problem': ndim_problem,
               'lower_boundary': -600*np.ones((ndim_problem,)),
               'upper_boundary': 600*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 301850,
               'n_individuals': 750,
               'seed_rng': 0}  # undefined in the original paper
    emna = EMNA(problem, options)
    results = emna.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 1.2459032377053032 vs 8.7673e-06 (from the original paper)



"""Repeat the following paper for `COEA`:
    Potter, M.A. and De Jong, K.A., 1994, October.
    A cooperative coevolutionary approach to function optimization.
    In International Conference on Parallel Problem Solving from Nature (pp. 249-257).
    Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007/3-540-58484-6_269

    All generated figures can be accessed via the following link:
    https://github.com/Evolutionary-Intelligence/pypop/tree/main/docs/repeatability/coea

    The original paper used the binary representation, which is *rarely* used for continuous optimization.
    Our Python implementation uses the real-valued representation, with a different sub-optimizer.

    Luckily our code could still repeat the data reported in the original paper *well*.
    Therefore, we argue that its repeatability could be **well-documented**.
"""
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.benchmarks.base_functions import rastrigin, griewank, ackley
from pypop7.optimizers.cc.coea import COEA as Solver


if __name__ == '__main__':
    sns.set_theme(style='darkgrid')

    ndim_problem = 10
    problem = {'fitness_function': griewank,
               'ndim_problem': ndim_problem,
               'lower_boundary': -600*np.ones((ndim_problem,)),
               'upper_boundary': 600*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 100000,
               'seed_rng': 0,
               'saving_fitness': 1}
    solver = Solver(problem, options)
    results = solver.optimize()
    print(results)
    fitness = results['fitness']
    plt.plot(fitness[:, 0], fitness[:, 1], 'k')
    plt.xticks([0, 20000, 40000, 60000, 80000, 100000])
    plt.xlim([0, 100000])
    plt.xlabel('function evaluations')
    plt.yticks([0, 2, 4, 6, 8])
    plt.ylim([0, 8])
    plt.ylabel('best individual')
    plt.show()

    ndim_problem = 20
    problem = {'fitness_function': rastrigin,
               'ndim_problem': ndim_problem,
               'lower_boundary': -5*np.ones((ndim_problem,)),
               'upper_boundary': 5*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 100000,
               'seed_rng': 0,
               'saving_fitness': 1}
    solver = Solver(problem, options)
    results = solver.optimize()
    print(results)
    fitness = results['fitness']
    plt.plot(fitness[:, 0], fitness[:, 1], 'k')
    plt.xticks([0, 20000, 40000, 60000, 80000, 100000])
    plt.xlim([0, 100000])
    plt.xlabel('function evaluations')
    plt.yticks([0, 10, 20, 30, 40])
    plt.ylim([0, 40])
    plt.ylabel('best individual')
    plt.show()

    ndim_problem = 30
    problem = {'fitness_function': ackley,
               'ndim_problem': ndim_problem,
               'lower_boundary': -600*np.ones((ndim_problem,)),
               'upper_boundary': 600*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 100000,
               'seed_rng': 0,
               'saving_fitness': 1}
    solver = Solver(problem, options)
    results = solver.optimize()
    print(results)
    fitness = results['fitness']
    plt.plot(fitness[:, 0], fitness[:, 1], 'k')
    plt.xticks([0, 20000, 40000, 60000, 80000, 100000])
    plt.xlim([0, 100000])
    plt.xlabel('function evaluations')
    plt.yticks([0, 4, 8, 12, 16])
    plt.ylim([0, 16])
    plt.ylabel('best individual')
    plt.show()



"""Repeat the following paper for `LAMCTS`:
    Wang, L., Fonseca, R. and Tian, Y., 2020.
    Learning search space partition for black-box optimization using monte carlo tree search.
    Advances in Neural Information Processing Systems, 33, pp.19511-19522.
    https://arxiv.org/abs/2007.00708 (an updated version)
    https://proceedings.neurips.cc/paper/2020/hash/e2ce14e81dba66dbff9cbc35ecfdb704-Abstract.html
    (the original version)

    Luckily our Python code could still repeat the data reported in the updated Pyton code *well*.
    Therefore, we argue that its repeatability could be **well-documented**.
"""
import time

import numpy as np

from pypop7.optimizers.bo.lamcts import LAMCTS as Solver


class Levy(object):
    def __init__(self, ndim):
        self.ndim = ndim

    def __call__(self, x):
        w = 1.0 + (x - 1.0)/4.0
        y = np.sin(np.pi*w[0])**2 + np.sum((w[1:self.ndim - 1] - 1.0)**2 * (
                1.0 + 10.0*np.sin(np.pi*w[1:self.ndim - 1] + 1.0)**2)) + (
                w[self.ndim - 1] - 1.0)**2 * (1.0 + np.sin(2.0*np.pi*w[self.ndim - 1])**2)
        return y


if __name__ == '__main__':
    start_run = time.time()
    ndim_problem = 100
    problem = {'fitness_function': Levy(ndim_problem),
               'ndim_problem': ndim_problem,
               'lower_boundary': -10*np.ones((ndim_problem,)),
               'upper_boundary': 10*np.ones((ndim_problem,))}
    options = {'seed_rng': 3,
               'verbose': 1,
               'saving_fitness': 10,
               'max_function_evaluations': 15000}
    solver = Solver(problem, options)
    results = solver.optimize()
    print(results)
    print('*** Runtime: {:7.5e}'.format(time.time() - start_run))
    # n_function_evaluations 15000: best_so_far_y 45.292



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer
from pypop7.optimizers.pso.pso import PSO


class IPSO(PSO):
    """Incremental Particle Swarm Optimizer (IPSO).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'n_individuals' - swarm (population) size, aka number of particles (`int`, default: `20`),
                * 'constriction'  - constriction factor (`float`, default: `0.729`),
                * 'cognition'     - cognitive learning rate (`float`, default: `2.05`),
                * 'society'       - social learning rate (`float`, default: `2.05`),
                * 'max_ratio_v'   - maximal ratio of velocities w.r.t. search range (`float`, default: `0.5`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.pso.ipso import IPSO
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022}
       >>> ipso = IPSO(problem, options)  # initialize the optimizer class
       >>> results = ipso.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"IPSO: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       IPSO: 5000, 2.29225104244031e-07

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/4pk3ssrf>`_ for more details.

    Attributes
    ----------
    cognition     : `float`
                    cognitive learning rate, aka acceleration coefficient.
    constriction  : `float`
                    constriction factor.
    max_ratio_v   : `float`
                    maximal ratio of velocities w.r.t. search range.
    n_individuals : `int`
                    swarm (population) size, aka number of particles.
    society       : `float`
                    social learning rate, aka acceleration coefficient.

    References
    ----------
    De Oca, M.A.M., Stutzle, T., Van den Enden, K. and Dorigo, M., 2011.
    Incremental social learning in particle swarms.
    IEEE Transactions on Systems, Man, and Cybernetics, Part B (Cybernetics), 41(2), pp.368-384.
    https://ieeexplore.ieee.org/document/5582312
    """
    def __init__(self, problem, options):
        PSO.__init__(self, problem, options)
        self.n_individuals = 1  # minimum of swarm size
        self.max_n_individuals = options.get('max_n_individuals', 1000)  # maximum of swarm size
        assert self.max_n_individuals > 0
        self.cognition = options.get('cognition', 2.05)  # cognitive learning rate
        assert self.cognition > 0.0
        self.society = options.get('society', 2.05)  # social learning rate
        assert self.society > 0.0
        self.constriction = options.get('constriction', 0.729)  # constriction factor
        assert self.constriction > 0.0
        self.max_ratio_v = options.get('max_ratio_v', 0.5)  # maximal ratio of velocity
        assert 0.0 <= self.max_ratio_v <= 1.0

    def initialize(self, args=None):
        v = np.zeros((self.n_individuals, self.ndim_problem))  # velocities
        x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary,
                                            size=self._swarm_shape)  # positions
        y = np.empty((self.n_individuals,))  # fitness
        p_x, p_y = np.copy(x), np.copy(y)  # personally previous-best positions and fitness
        for i in range(self.n_individuals):
            if self._check_terminations():
                return v, x, y, p_x, p_y
            y[i] = self._evaluate_fitness(x[i], args)
        p_y = np.copy(y)
        return v, x, y, p_x, p_y

    def iterate(self, v=None, x=None, y=None, p_x=None, p_y=None, args=None, fitness=None):
        for i in range(self.n_individuals):  # horizontal social learning
            if self._check_terminations():
                return v, x, y, p_x, p_y
            cognition_rand = self.rng_optimization.uniform(size=(self.ndim_problem,))
            society_rand = self.rng_optimization.uniform(size=(self.ndim_problem,))
            v[i] = self.constriction*(v[i] + self.cognition*cognition_rand*(p_x[i] - x[i]) +
                                      self.society*society_rand*(p_x[np.argmin(p_y)] - x[i]))  # velocity update
            v[i] = np.clip(v[i], self._min_v, self._max_v)
            x[i] += v[i]  # position update
            x[i] = np.clip(x[i], self.lower_boundary, self.upper_boundary)
            y[i] = self._evaluate_fitness(x[i], args)
            if y[i] < p_y[i]:  # online update
                p_x[i], p_y[i] = x[i], y[i]
        if self.n_individuals < self.max_n_individuals:  # population growth (vertical social learning)
            if self._check_terminations():
                return v, x, y, p_x, p_y
            xx = self.rng_optimization.uniform(self.lower_boundary, self.upper_boundary)
            model = p_x[np.argmin(p_y)]  # the best particle is used as model
            # use different random numbers of different dimensions for diversity (important),
            # which is *slightly different* from the original paper but often with better performance
            # xx += self.rng_optimization.uniform()*(model - xx)  # from the original paper
            xx += self.rng_optimization.uniform(size=(self.ndim_problem,))*(model - xx)
            xx = np.clip(xx, self.lower_boundary, self.upper_boundary)
            yy = self._evaluate_fitness(xx, args)
            v = np.vstack((v, np.zeros((self.ndim_problem,))))
            x, y = np.vstack((x, xx)), np.hstack((y, yy))
            p_x, p_y = np.vstack((p_x, xx)), np.hstack((p_y, yy))
            self.n_individuals += 1
        self._n_generations += 1
        return v, x, y, p_x, p_y

    def optimize(self, fitness_function=None, args=None):
        fitness = Optimizer.optimize(self, fitness_function)
        v, x, y, p_x, p_y = self.initialize(args)
        while not self.termination_signal:
            self._print_verbose_info(fitness, y)
            v, x, y, p_x, p_y = self.iterate(v, x, y, p_x, p_y, args)
        return self._collect(fitness, y)



"""Repeat the following paper for `MAES`:
    Beyer, H.G. and Sendhoff, B., 2017.
    Simplify your covariance matrix adaptation evolution strategy.
    IEEE Transactions on Evolutionary Computation, 21(5), pp.746-759.
    https://ieeexplore.ieee.org/document/7875115
    https://homepages.fhv.at/hgb/downloads/ForDistributionFastMAES.tar

    All generated figures can be accessed via the following link:
    https://github.com/Evolutionary-Intelligence/pypop/tree/main/docs/repeatability/maes

    Luckily our Python code could repeat the data reported in the original paper *well*.
    Therefore, we argue that its repeatability could be **well-documented**.
"""
import pickle

import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.benchmarks.base_functions import sphere, cigar, discus, ellipsoid
from pypop7.optimizers.es.es import ES
from pypop7.optimizers.es.maes import MAES


def plot(function, ndim):
    plt.figure()
    result = pickle.load(open(function + '_' + str(ndim) + '.pickle', 'rb'))
    plt.yscale('log')
    plt.plot(np.arange(len(results['f'])), results['f'], 'r-')
    plt.plot(np.arange(len(results['stepsize'])), results['stepsize'], 'm-')
    if function == 'sphere':
        plt.ylim([1e-15, 1e2])
        plt.yticks([1e-14, 1e-12, 1e-10, 1e-8, 1e-6, 1e-4, 1e-2, 1e0, 1e2])
        if ndim == 3:
            plt.xlim([0, 150])
            plt.xticks([0, 20, 40, 60, 80, 100, 120, 140])
        elif ndim == 30:
            plt.xlim([0, 600])
            plt.xticks([0, 100, 200, 300, 400, 500, 600])
    elif function == 'cigar':
        plt.ylim([1e-15, 1e8])
        plt.yticks([1e-15, 1e-10, 1e-5, 1e0, 1e5])
        if ndim == 3:
            plt.xlim([0, 250])
            plt.xticks([0, 50, 100, 150, 200, 250])
        elif ndim == 30:
            plt.xlim([0, 1400])
            plt.xticks([0, 200, 400, 600, 800, 1000, 1200, 1400])
    elif function == 'discus':
        plt.ylim([1e-15, 1e6])
        plt.yticks([1e-15, 1e-10, 1e-5, 1e0, 1e5])
        if ndim == 3:
            plt.xlim([0, 280])
            plt.xticks([0, 50, 100, 150, 200, 250])
        elif ndim == 30:
            plt.xlim([0, 3000])
            plt.xticks([0, 500, 1000, 1500, 2000, 2500])
    elif function == 'ellipsoid':
        plt.ylim([1e-15, 1e6])
        plt.yticks([1e-15, 1e-10, 1e-5, 1e0, 1e5])
        if ndim == 3:
            plt.xlim([0, 250])
            plt.xticks([0, 50, 100, 150, 200, 250])
        elif ndim == 30:
            plt.xlim([0, 3000])
            plt.xticks([0, 500, 1000, 1500, 2000, 2500, 3000])
    plt.xlabel("g")
    plt.ylabel(r"$\sigma^{(g)}$  and  $f(y^{(g)})$")
    plt.title(function.capitalize() + " N = " + str(ndim))
    plt.show()


class Fig4(MAES):
    def optimize(self, fitness_function=None, args=None):
        fitness = ES.optimize(self, fitness_function)
        z, d, mean, s, tm, y = self.initialize()
        best_f, stepsize = [], []  # for plotting data
        while not self._check_terminations():
            # sample and evaluate offspring population
            z, d, y = self.iterate(z, d, mean, tm, y, args)
            mean, s, tm = self._update_distribution(z, d, mean, s, tm, y)
            self._print_verbose_info(fitness, y)
            self._n_generations += 1
            best_f.append(np.min(y))
            stepsize.append(self.sigma)
            if self.is_restart:
                z, d, mean, s, tm, y = self.restart_reinitialize(z, d, mean, s, tm, y)
        res = self._collect(fitness, y, mean)
        res['s'] = s
        res['f'] = best_f
        res['stepsize'] = stepsize
        # by default, do NOT save transformation matrix of search distribution in order to save memory,
        # owing to its *quadratic* space complexity
        return res


if __name__ == '__main__':
    sns.set_theme(style='darkgrid')
    for f in [sphere, cigar, discus, ellipsoid]:
        for dim in [3, 30]:
            problem = {'fitness_function': f,
                       'ndim_problem': dim}
            options = {'fitness_threshold': 1e-15,
                       'seed_rng': 2022,  # not given in the original paper
                       'x': np.ones((dim,)),
                       'sigma': 1.0,
                       'saving_fitness': 1,
                       'is_restart': False}
            solver = Fig4(problem, options)
            results = solver.optimize()
            pickle.dump(results, open(f.__name__ + '_' + str(dim) + '.pickle', 'wb'))
            plot(f.__name__, dim)



import numpy as np

from pypop7.optimizers.es.es import ES
from pypop7.optimizers.es.opoc2006 import OPOC2006


class OPOC2009(OPOC2006):
    """(1+1)-Cholesky-CMA-ES 2009 (OPOC2009).

    Parameters
    ----------
    problem : `dict`
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : `dict`
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`).
              and with the following particular settings (`keys`):
                * 'sigma' - initial global step-size, aka mutation strength (`float`),
                * 'mean'  - initial (starting) point, aka mean of Gaussian search distribution (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

    Examples
    --------
    Use the optimizer `OPOC2009` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.opoc2009 import OPOC2009
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3*numpy.ones((2,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> opoc2009 = OPOC2009(problem, options)  # initialize the optimizer class
       >>> results = opoc2009.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"OPOC2009: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       OPOC2009: 5000, 2.7686802211556655e-17

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/3ba4ctny>`_ for more details.

    References
    ----------
    Suttorp, T., Hansen, N. and Igel, C., 2009.
    Efficient covariance matrix update for variable metric evolution strategies.
    Machine Learning, 75(2), pp.167-197.
    https://link.springer.com/article/10.1007/s10994-009-5102-1
    (See Algorithm 2 for details.)
    """
    def __init__(self, problem, options):
        OPOC2006.__init__(self, problem, options)

    def initialize(self, args=None, is_restart=False):
        mean, y, a, best_so_far_y, p_s = OPOC2006.initialize(self, args, is_restart)
        p_c = np.zeros((self.ndim_problem,))  # evolution path
        a_i = np.diag(np.ones((self.ndim_problem,)))  # inverse of Cholesky factors
        return mean, y, a, a_i, best_so_far_y, p_s, p_c

    def iterate(self, mean=None, a=None, a_i=None, best_so_far_y=None, p_s=None, p_c=None, args=None):
        # sample and evaluate only one offspring
        z = self.rng_optimization.standard_normal((self.ndim_problem,))
        x = mean + self.sigma*np.dot(a, z)
        y = self._evaluate_fitness(x, args)
        l_s = 1 if y <= best_so_far_y else 0
        p_s = (1.0 - self.c_p)*p_s + self.c_p*l_s
        self.sigma *= np.exp(self.lr_sigma*(p_s - self.p_ts)/(1.0 - self.p_ts))
        if y <= best_so_far_y:
            mean, best_so_far_y = x, y
            if p_s < self.p_t:
                p_c = (1.0 - self.c_c)*p_c + np.sqrt(self.c_c*(2.0 - self.c_c))*np.dot(a, z)
                alpha = 1.0 - self.c_cov
            else:
                p_c *= 1.0 - self.c_c
                alpha = 1.0 - self.c_cov + self.c_cov*self.c_c*(2.0 - self.c_c)
            beta = self.c_cov
            w = np.dot(a_i, p_c)
            w_norm = np.power(np.linalg.norm(w), 2)
            s_w_norm = np.sqrt(1.0 + beta/alpha*w_norm)
            a = np.sqrt(alpha)*a + np.sqrt(alpha)/w_norm*(s_w_norm - 1.0)*np.dot(
                p_c[:, np.newaxis], w[np.newaxis, :])
            a_i = 1.0/np.sqrt(alpha)*a_i - 1.0/(np.sqrt(alpha)*w_norm)*(1.0 - 1.0/s_w_norm)*np.dot(
                w[:, np.newaxis], np.dot(w[np.newaxis, :], a_i))
        return mean, y, a, a_i, best_so_far_y, p_s, p_c

    def restart_reinitialize(self, mean=None, y=None, a=None, a_i=None, best_so_far_y=None,
                             p_s=None, p_c=None, fitness=None, args=None):
        self._list_fitness.append(best_so_far_y)
        is_restart_1, is_restart_2 = self.sigma < self.sigma_threshold, False
        if len(self._list_fitness) >= self.stagnation:
            is_restart_2 = (self._list_fitness[-self.stagnation] - self._list_fitness[-1]) < self.fitness_diff
        is_restart = bool(is_restart_1) or bool(is_restart_2)
        if is_restart:
            self._print_verbose_info(fitness, y, True)
            if self.verbose:
                print(' ....... *** restart *** .......')
            self._n_restart += 1
            self._list_generations.append(self._n_generations)  # for each restart
            self._n_generations = 0
            self.sigma = np.copy(self._sigma_bak)
            mean, y, a, a_i, best_so_far_y, p_s, p_c = self.initialize(args, True)
            self._list_fitness = [best_so_far_y]
        return mean, y, a, a_i, best_so_far_y, p_s, p_c

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        mean, y, a, a_i, best_so_far_y, p_s, p_c = self.initialize(args)
        while not self._check_terminations():
            self._print_verbose_info(fitness, y)
            mean, y, a, a_i, best_so_far_y, p_s, p_c = self.iterate(
                mean, a, a_i, best_so_far_y, p_s, p_c, args)
            self._n_generations += 1
            if self.is_restart:
                mean, y, a, a_i, best_so_far_y, p_s, p_c = self.restart_reinitialize(
                    mean, y, a, a_i, best_so_far_y, p_s, p_c, fitness, args)
        return self._collect(fitness, y, mean)
